import inspect
import sklearn
import sys

from msdss_models_api.models import Model
from sklearn import *

from .defaults import *
from .tools import *

def get_sklearn_models(base_module=__name__, modules=DEFAULT_MODULES):
    """
    Extract scikit-learn models and convert them into :class:`msdss_models_api:msdss_models_api.models.Model` classes with appropriate methods for input, output, and update.

    To extract scikit-learn models, the general process is:

    1. Get the models module (modules are manually defined from inspecting the docs)
    2. Filter out all names in the module where the first letter is not capitalized or contains an underscore
    3. Filter out all names in the module that are not classes or are not part of the module
    4. Dynamically create a class for the resulting names, extending from :class:`msdss_models_api:msdss_models_api.models.Model`
    5. Add methods:
        
        * ``input`` using :func:`msdss_models_sklearn.tools.create_input_method`
        * ``output`` using :func:`msdss_models_sklearn.tools.create_output_method`
        * ``update`` using :func:`msdss_models_sklearn.tools.create_update_method`
    
    Parameters
    ----------
    base_module : str
        Name of the base module to attach model namespaces to.
    modules : list(str)
        List of scikit-learn module names to extract models from.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. jupyter-execute::

        from msdss_models_sklearn import models as sklearn_models
        from pprint import pprint

        # Print all converted models
        pprint(sklearn_models)
    """
    out = {}
    for name in modules:

        # (get_sklearn_models_module) Get the sklearn module
        module = getattr(sklearn, name)

        # (get_sklearn_models_filter1) Filter for first letter caps and no underscore in module namespace
        module_models = [getattr(module, m) for m in dir(module) if m[0] == m[0].upper() and m[0] != '_']

        # (get_sklearn_models_filter2) Filter for classes that are inside the module only
        module_models = [m for m in module_models if inspect.isclass(m) and m.__module__.startswith(f'sklearn.{name}')]

        # (get_sklearn_models_create) Create msdss model and add to dict
        for model in module_models:
            model_name = f'sklearn.{name}.{model.__name__}'
            out[model_name] = type(model.__name__, (Model,), {
                'input': create_input_method(model),
                'output': create_output_method(),
                'update': create_update_method()
            })
            setattr(sys.modules[base_module], model.__name__, out[model_name])
    return out

models = get_sklearn_models(__name__)


DEFAULT_MODULES = [
    'calibration',
    'cluster',
    'compose',
    'covariance',
    'cross_decomposition',
    'decomposition',
    'discriminant_analysis',
    'ensemble',
    'feature_extraction',
    'feature_selection',
    'gaussian_process',
    'impute',
    'isotonic',
    'kernel_approximation',
    'kernel_ridge',
    'linear_model',
    'manifold',
    'mixture',
    'naive_bayes',
    'neighbors',
    'neural_network',
    'preprocessing',
    'random_projection',
    'semi_supervised',
    'svm',
    'tree'
]


import pandas

def create_input_method(model):
    """
    Create model input method for scikit-learn models to be compatible with :class:`msdss_models_api:msdss_models_api.models.Model`.

    See :class:`msdss_models_api:msdss_models_api.models.Model.input`.

    Parameters
    ----------
    model : class
        Scikit-learn model class to create machine learning models.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. jupyter-execute::

        from msdss_models_sklearn.tools import *
        from sklearn.linear_model import LinearRegression

        input = create_input_method(LinearRegression)
    """
    def input(self, data, x=None, y=None, _fit={}, *args, **kwargs):
        
        # (create_input_method_vars) Set default vars
        x = x if x else self.settings['x'] if 'x' in self.settings else x
        y = y if y else self.settings['y'] if 'y' in self.settings else y

        # (create_input_method_data) Format data for model instance input
        data = pandas.DataFrame(data)
        data_x = data[x] if x else data
        data_y = data[y] if y else None

        # (create_input_method_set) Train model instance
        self.instance = model(*args, **kwargs).fit(data_x, data_y, **_fit)
    return input

def create_output_method():
    """
    Create model output method for scikit-learn models to be compatible with :class:`msdss_models_api:msdss_models_api.models.Model`.

    See :class:`msdss_models_api:msdss_models_api.models.Model.output`.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. jupyter-execute::

        from msdss_models_sklearn.tools import *

        input = create_output_method()
    """
    def output(self, data, x=None, y=None, *args, **kwargs):

        # (create_output_method_vars) Set default vars
        x = x if x else self.settings['x'] if 'x' in self.settings else x
        y = y if y else self.settings['y'] if 'y' in self.settings else y
        y = [y] if y and not isinstance(y, list) else y

        # (create_output_method_data) Format data for model instance output
        data = pandas.DataFrame(data)
        data_x = data[x] if x else data

        # (create_output_method_output) Get output from trained model instance
        model = self.instance
        if 'predict' in dir(model):
            out = pandas.DataFrame(model.predict(data_x, *args, **kwargs))
        else:
            out = pandas.DataFrame(model.transform(data_x, *args, **kwargs))
        
        # (create_output_method_return) Set column names if avail and return output
        if y:
            out.columns = y
        return out
    return output

def create_update_method():
    """
    Create model update method for scikit-learn models to be compatible with :class:`msdss_models_api:msdss_models_api.models.Model`.

    See :class:`msdss_models_api:msdss_models_api.models.Model.update`.

    Author
    ------
    Richard Wen <rrwen.dev@gmail.com>
    
    Example
    -------
    .. jupyter-execute::

        from msdss_models_sklearn.tools import *

        input = create_output_method()
    """
    def update(self, data, x=None, y=None, *args, **kwargs):

        # (create_update_method_vars) Set default vars
        x = x if x else self.settings['x'] if 'x' in self.settings else x
        y = y if y else self.settings['y'] if 'y' in self.settings else y

        # (create_update_method_data) Format data for update
        data = pandas.DataFrame(data)
        data_x = data[x] if x else data
        data_y = data[y] if y else None

        # (create_update_method_update) Update model instance
        self.instance.fit(data_x, data_y, *args, **kwargs)
    return update
